from random import shuffle as rshuffle

class Stack():
    def __init__(self):
        self.__stk = []
    def __len__(self):
        return len(self.__stk)
    def __str_rep(self, item):
        if type(item) in (int, float):
            return str(item)
        elif type(item) is str:
            if len(item) == 1:
                return "'"+item
            else:
                return '"'+item+'"'
        elif type(item) is bool:
            return ['false', 'true'][item]
        elif type(item) in (list, tuple):
            return '[' + ' '.join(list(map(self.__str_rep, item))) + ']'
        elif type(item) is set:
            return '{' + ' '.join(list(map(self.__str_rep, item))) + '}'
    def __str__(self):
        return '['+' '.join(list(map(self.__str_rep, self.__stk))) + ']'
    def pop(self, index=-1):
        return self.__stk.pop(index)
    def push(self, item, index=-1):
        if index == -1: index = len(self)
        self.__stk.insert(index, item)
    def rotate(self, moves=1):
        if moves < 1:
            moves = abs(moves) % len(self.__stk)
            if moves > 0:
                self.__stk = self.__stk[moves:] + self.__stk[:moves]
        else:
            moves = -moves
            self.__stk = self.__stk[moves:] + self.__stk[:moves]
    def item_at(self, index):
        return self.__stk[index]
    def index_of(self, item):
        return self.__stk.index(item)
    def peek(self):
        return self.__stk[-1]
    def reverse(self):
        self.__stk = self.__stk[::-1]
    def shuffle(self):
        rshuffle(self.__stk)
    def sort(self, asc = True):
        self.__stk = sorted(self.__stk)[::2*asc - 1]



import os
import argparse
from .stack import *
from .chords import all_chords

def trycast(token):
    try:
        return int(token)
    except ValueError:
        pass
    try:
        return float(token)
    except ValueError:
        pass
    return None

def process(tokens, past_stack=None):
    stack = past_stack
    if not stack:
        stack = Stack()
    index = 0
    while 0 <= index < len(tokens):
        if tokens[index] in all_chords.keys():
            all_chords[tokens[index]](stack)
        if tokens[index].startswith('|'):
            st = tokens[index][1:]
            if st.endswith('|') and not st.endswith(r'\|'):
                st = st[:-1]
            else:
                index += 1
                while index < len(tokens):
                    st += " " + tokens[index]
                    if tokens[index].endswith('|') and not tokens[index].endswith(r'\|'):
                        st = st[:-1]
                        break
                    elif tokens[index].endswith(r'\|'):
                        st = st[:-2] + "|"
                    index += 1
            stack.push(st)
        else:
            casted = trycast(tokens[index])
            if casted is not None:
                stack.push(casted)
        index += 1
    print(stack)
    return stack

def readf(path, lyrics):
    prog = ''
    with open(path, 'r') as file:
        prog = file.read()
    prog = prog.split('\n')
    if lyrics:
        prog = prog[::2]
    prog = [x.split() for x in prog]
    tokens = []
    for row in prog:
        tokens += row
    process(tokens)

def replenv():
    inp = input("td> ")
    cstack = Stack()
    while inp != "quit":
        cstack = process(inp.split(), cstack)
        inp = input("td> ")

def main():
    parser = argparse.ArgumentParser(description = 'Interpreter for the tone-deaf programming language.')
    parser.add_argument('-f', type=str, help='Open a tone-deaf source file.')
    parser.add_argument('-l', action='store_true', help='Open the source file in "lyric mode".')
    args = parser.parse_args()
    if args.f != None:
        if args.f != '':
            if os.path.isfile(args.f):
                readf(args.f, args.l)
            else:
                print('-f: Error - no path given.')
        else:
            print('-f: Error - no path given.')
    else:
        replenv()

if __name__ == '__main__':
    main()



# D Chords are for string / list / set manipulation

from random import shuffle

def reverse_str(stack):
    s = stack.pop()
    stack.push(s[::-1])

def shuffle_str(stack):
    s = stack.pop()
    if isinstance(s, str):
        s = list(s)
        shuffle(s)
        s = ''.join(s)
    else:
        shuffle(s)
    stack.push(s)

def slice_str(stack):
    end = stack.pop()
    start = stack.pop()
    s = stack.pop()
    end = len(s) if end == 0 else end
    start = 0 if start == 0 else start
    stack.push(s[start:end])

def slice_str_adv(stack):
    skip = stack.pop()
    end = stack.pop()
    start = stack.pop()
    s = stack.pop()
    skip = 1 if skip == 0 else skip
    end = len(s) if end == 0 else end
    start = 0 if start == 0 else start
    stack.push(s[start:end:skip])

def keep_every_other(stack):
    s = stack.pop()
    if isinstance(s, int):
        v = stack.pop()
        stack.push(v[s::2])
    else:
        stack.push(s[::2])

def keep_every_third(stack):
    s = stack.pop()
    if isinstance(s, int):
        v = stack.pop()
        stack.push(v[s::3])
    else:
        stack.push(s[::3])

def keep_every_fifth(stack):
    s = stack.pop()
    if isinstance(s, int):
        v = stack.pop()
        stack.push(v[s::5])
    else:
        stack.push(s[::5])

def cast_to_set(stack):
    s = stack.pop()
    stack.push(set(s))

def palindromize(stack):
    s = stack.pop()
    stack.push(s + s[::-1][1:])

def is_palindrome(stack):
    s = stack.pop()
    stack.push(s == s[::-1])

chords = {
    'D'    : reverse_str,
    'D#'   : shuffle_str,
    'D#4'  : slice_str,
    'D#7'  : slice_str_adv,
    'D#m'  : keep_every_other,
    'D#m7' : keep_every_third,
    'D#M7' : keep_every_fifth,
    'D^9'  : cast_to_set,
    'Dm7'  : palindromize,
    'Dm9'  : is_palindrome
}

__all__ = [ "chords" ]



from .a_chords import chords as a_chords
from .d_chords import chords as d_chords
from .e_chords import chords as e_chords

all_chords = dict()
all_chords.update(a_chords)
all_chords.update(d_chords)
all_chords.update(e_chords)

__all__ = [ "all_chords" ]



from setuptools import setup, find_packages

import tone_deaf

long_desc = ""
try:
    import pypandoc
    long_desc = pypandoc.convert('README.md', 'rst', extra_args = ('--eol', 'lf'))
except(IOError, ImportError):
    long_desc = open('README.md').read()

setup(
    name = "tone-deaf",
    version = "0.3.0",
    description = "An interpreter for the tone-deaf programming language.",
    long_description = long_desc,
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6"
    ],
    entry_points = {
        'console_scripts': [
            'tone-deaf = tone_deaf:main'
        ]
    },
    keywords = "tone deaf esoteric programming language interpreter",
    author = "Kade Robertson",
    author_email = "kade@kaderobertson.pw",
    url = "https://github.com/kade-robertson/tone-deaf",
    license = "MIT",
    packages = find_packages(),
    install_requires = [],
    python_requires = '>=3.4, <4'
)



# Generic functions that might be used in multiple chords.

def tetrate(x, y):
    out = x
    while y > 1:
        out = x ** out
        y -= 1
    return out

def is_prime(x):
    if x % 2 == 0 or x < 2:
        return 0
    elif x == 2:
        return 1
    else:
        i = 3
        while i < int(x**.5)+1:
            if x % i == 0:
                return 0
            i += 2
        return 1

def prime_factors(x):
    f = []
    while x % 2 == 0:
        f += [2]
        x /= 2
    i = 3
    while x > 1:
        while x % i == 0:
            f += [i]
            x /= i
        i += 2
    return f

def tetr_mod(x, y, m):
    a = 1
    i = 0
    while i < y:
        b = pow(x, a, m)
        if a == b: break
        a = b
        i += 1
    return b

def lu_decomp_det(m):
    d = 1
    s = len(m)
    for i in range(s):
        d *= m[i,i]
        c = i+1
        for j in range(c, s):
            m[j,i] /= m[i,i]
            for k in range(c, s):
                m[j,k] -= m[j,i] * m[i,k]
    return d



# A Chords are math operations.

import math
import fractions
from .utils import *

def pop_2_and_add(stack):
    stack.push(stack.pop() + stack.pop())

def pop_2_and_subtr(stack):
    a = stack.pop()
    stack.push(stack.pop() - a)

def pop_2_and_mult(stack):
    stack.push(stack.pop() * stack.pop())

def pop_2_and_divd(stack):
    a = stack.pop()
    stack.push(stack.pop() / a)

def pop_2_and_modul(stack):
    a = stack.pop()
    stack.push(stack.pop() % a)

def pop_and_square(stack):
    a = stack.pop()
    stack.push(a * a)

def pop_and_cube(stack):
    a = stack.pop()
    stack.push(a * a * a)

def pop_and_sqrt(stack):
    stack.push(stack.pop() ** 0.5)

def pop_and_cbrt(stack):
    stack.push(stack.pop() ** (1/3))

def pop_2_and_exp(stack):
    a = stack.pop()
    stack.push(stack.pop() ** a)

def pop_2_and_tetr(stack):
    a = stack.pop()
    stack.push(tetrate(stack.pop(), a))

def pop_2_and_check_greater(stack):
    a = stack.pop()
    stack.push(1 if stack.pop() > a else 0)

def pop_2_and_check_greq(stack):
    a = stack.pop()
    stack.push(1 if stack.pop() >= a else 0)

def pop_and_check_parity(stack):
    stack.push(stack.pop()%2)

def pop_and_check_primality(stack):
    stack.push(is_prime(stack.pop()))

def pop_and_get_prime_factors(stack):
    stack.push(prime_factors(stack.pop()))

def pop_3_and_exp_mod(stack):
    a = stack.pop()
    b = stack.pop()
    stack.push(pow(stack.pop(), b, a))

def pop_3_and_tetr_mod(stack):
    a = stack.pop()
    b = stack.pop()
    stack.push(tetr_mod(stack.pop(), b, a))

def pop_and_double(stack):
    stack.push(2 * stack.pop())

def pop_and_halve(stack):
    stack.push(stack.pop() / 2)

def pop_and_round(stack):
    stack.push(round(stack.pop()))

def pop_and_floor(stack):
    stack.push(math.floor(stack.pop()))

def pop_and_ceil(stack):
    stack.push(math.ceil(stack.pop()))

def pop_2_and_divd_int(stack):
    a = stack.pop()
    stack.push(stack.pop() // a)

def pop_2_and_divmod(stack):
    a = stack.pop()
    b = stack.pop()
    stack.push([b // a, b % a])

def pop_2_and_get_gcd(stack):
    a = stack.pop()
    stack.push(fractions.gcd(stack.pop(), a))

def pop_2_and_get_lcm(stack):
    a = stack.pop()
    b = stack.pop()
    stack.push((a*b)/fractions.gcd(a,b))

def pop_and_get_nth_trinum(stack):
    a = stack.pop()
    stack.push((a*-~a)/2)

def pop_2_and_get_perm(stack):
    a = stack.pop()
    b = stack.pop()
    stack.push(math.factorial(b)/math.factorial(b-a))

def pop_2_and_get_comb(stack):
    a = stack.pop()
    b = stack.pop()
    stack.push(math.factorial(b)/(math.factorial(a)*math.factorial(b-a)))

chords = {
    'A'    : pop_2_and_add,
    'As'   : pop_2_and_subtr,
    'Am'   : pop_2_and_mult,
    'Ad'   : pop_2_and_divd,
    'Ab'   : pop_2_and_modul,
    'A+'   : pop_and_square,
    'A#'   : pop_and_cube,
    'A4'   : pop_and_sqrt,
    'A6'   : pop_and_cbrt,
    'A7'   : pop_2_and_exp,
    'A9'   : pop_2_and_tetr,
    'A#4'  : pop_2_and_check_greater,
    'A#7'  : pop_2_and_check_greq,
    'A#d'  : pop_and_check_parity,
    'A#m'  : pop_and_check_primality,
    'A/D'  : pop_and_get_prime_factors,
    'A11'  : pop_3_and_exp_mod,
    'A13'  : pop_3_and_tetr_mod,
    'A7+'  : pop_and_double,
    'AM7'  : pop_and_halve,
    'Ab+'  : pop_and_round,
    'Ab4'  : pop_and_floor,
    'Ab7'  : pop_and_ceil,
    'Abd'  : pop_2_and_divd_int,
    'Abm'  : pop_2_and_divmod,
    'Am6'  : pop_2_and_get_gcd,
    'Am7'  : pop_2_and_get_lcm,
    'Am9'  : pop_and_get_nth_trinum,
    'A#M7' : pop_2_and_get_perm,
    'A#m7' : pop_2_and_get_comb
}

__all__ = [ "chords" ]



# E Chords perform stack manipulation.

import sys

def take_input(stack):
    stack.push(eval(input()))

def take_raw_input(stack):
    stack.push(input())

def take_raw_input_split_newlines(stack):
    stack.push(sys.stdin.read().split('\n'))

def take_raw_input_split_spaces(stack):
    stack.push(sys.stdin.read().split(' '))

def take_raw_input_split_delim(stack):
    stack.push(sys.stdin.read().split(str(stack.pop())))

def take_raw_input_split_comma(stack):
    stack.push(sys.stdin.read().split(','))

def pop_2_and_swap(stack):
    b = stack.pop()
    a = stack.pop()
    stack.push(b)
    stack.push(a)

def pop_and_discard(stack):
    stack.pop()

def pop_and_push_twice(stack):
    item = stack.pop()
    stack.push(item)
    stack.push(item)

def pop_and_push_thrice(stack):
    item = stack.pop()
    stack.push(item)
    stack.push(item)
    stack.push(item)

def rotate_stack_right(stack):
    stack.rotate(1)

def rotate_stack_left(stack):
    stack.rotate(-1)

def reverse_stack(stack):
    stack.reverse()

def shuffle_stack(stack):
    stack.shuffle()

def swap_first_and_third(stack):
    c = stack.pop()
    b = stack.pop()
    a = stack.pop()
    stack.push(c)
    stack.push(b)
    stack.push(a)

def pop_under_first(stack):
    stack.pop(-2)

def pop_two_under_first(stack):
    stack.pop(-3)
    stack.pop(-2)

def sort_stack(stack):
    stack.sort()

def sort_stack_desc(stack):
    stack.sort(False)

def swap_first_and_fourth(stack):
    d = stack.pop()
    c = stack.pop()
    b = stack.pop()
    a = stack.pop()
    stack.push(d)
    stack.push(b)
    stack.push(c)
    stack.push(a)

chords = { 
    'E'    : take_input,
    'Em'   : take_raw_input,
    'E5'   : take_raw_input_split_newlines,
    'E6'   : take_raw_input_split_spaces,
    'E7'   : take_raw_input_split_delim,
    'E9'   : take_raw_input_split_comma,
    'Es'   : pop_2_and_swap,
    'Em6'  : pop_and_discard,
    'Em7'  : pop_and_push_twice,
    'EM7'  : pop_and_push_thrice,
    'E11'  : rotate_stack_right,
    'Es4'  : rotate_stack_left,
    'E7#9' : reverse_stack,
    'E75b' : shuffle_stack,
    'E7b9' : swap_first_and_third,
    'Eb^9' : pop_under_first,
    'Em^9' : pop_two_under_first,
    'Ems4' : sort_stack,
    'Em/B' : sort_stack_desc,
    'Em/D' : swap_first_and_fourth
}

__all__ = [ "chords" ]
